/**
 * utility function to perform webhook signature verification
 * @function module:webhookUtil.verifyMessageFormat
 * @return {boolean} true if the webhook message received from Bots is verified successfully.
 * @param {string} signature - signature included in the bot message, to be compared to calculated signature.
 * @param {Buffer} msgBody - raw message body of the bot message.
 * @param {string} encoding - encoding of the raw message body.
 * @param {string} secretKey - secretKey used to calculate message signature
 */
export declare function verifyMessageFromBot(signature: any, msgBody: any, encoding: any, secretKey: any): boolean;
/**
 * utility function for use with expressjs route in handling the raw message body of the webhook message received from bot.
 * Instead of just letting bodyParser.json to parse the raw message to JSON, the rawMessage and its encoding is saved as properties
 * 'rawBody' and 'encoding' for use in signature verification in method verifyMessageFormat.
 * @function module:webhookUtil.bodyParserRawMessageVerify
 * @return {boolean} true if the webhook message received from Bots is verified successfully.
 * @param {object} req - expressjs req for the POST route.
 * @param {object} res - expressjs res for the POST route.
 * @param {Buffer} buf - the raw message body.
 * @param {string} encoding - encoding of the raw message body.
 */
export declare function bodyParserRawMessageVerify(req: any, res: any, buf: any, encoding: any): void;
/**
 * create the payload signature header.
 * @function module:Util/Webhook.buildSignatureHeader
 * @param {Buffer} buf - Raw payload as a Buffer, such as `Buffer.from(JSON.stringify(payload), 'utf8')`
 * @param {string} secret - secret key of the channel for computing signature
 * @param {string} [encoding] - secret key of the channel for computing signature
 */
declare function buildSignatureHeader(buf: any, secret: any, encoding?: any): string;
/**
 * utility function to send message to bot webhook channel, generating the right message with signature
 * @function module:webhookUtil.messageToBot
 * @param {string} channelUrl - send the message to this channel url
 * @param {string} channelSecretKey - secret key of the channel for computing message signature.
 * @param {string} userId - userId is the sender of the message.
 * @param {object|string} inMsg - message to be sent to bot
 * @param {function} callback - callback function to be invoked after message is sent
 */
export declare function messageToBot(channelUrl: any, channelSecretKey: any, userId: any, inMsg: any, callback: any): void;
/**
 * utility function to send message to bot webhook channel, generating the right message with signature.  This function also allows additional
 * properties to be sent along to the bot.  A common use case is to add a userProfile property.
 * @function module:webhookUtil.messageToBotWithProperties
 * @param {string} channelUrl - send the message to this channel url
 * @param {string} channelSecretKey - secret key of the channel for computing message signature.
 * @param {string} userId - userId is the sender of the message.
 * @param {object|string} inMsg - message to be sent to bot
 * @param {object} [additionalProperties] - additional properties like profile can be added
 * @param {function} callback - callback function to be invoked after message is sent
 */
export declare function messageToBotWithProperties(channelUrl: any, channelSecretKey: any, userId: any, inMsg: any, additionalProperties: any, callback: any): void;
/**
 * The webhookUtil is a set of utility functions for bot integration via webhook channel.
 * @module webhookUtil
 */
export declare const webhookUtil: {
    messageToBot: typeof messageToBot;
    messageToBotWithProperties: typeof messageToBotWithProperties;
    verifyMessageFromBot: typeof verifyMessageFromBot;
    bodyParserRawMessageVerify: typeof bodyParserRawMessageVerify;
    buildSignatureHeader: typeof buildSignatureHeader;
};
export {};
